from logging_tree import printout as print_logging_tree


def logging_tree(return_result = False, node=None):
    if return_result:
        try:
            from logging_tree.format import build_description as build_logging_tree_description
            return f'Logging tree:\n{build_logging_tree_description(node)[:-1]}'
        except BaseException as e:
            return f'Building logging tree failed. {e}'
    print_logging_tree(node)



import argparse
from .exceptions import CommandError, ConfigurationError
from .commands import commands
from logging import getLogger

logger = getLogger(__name__)


def setup_arg_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    for command in commands:
        command_parser = subparsers.add_parser(command.name, description=command.__doc__)
        command.initialize_parser(command_parser)
        command_parser.set_defaults(main=command.main)
    return parser


def main():
    parser = setup_arg_parser()
    args = parser.parse_args()
    try:
        if not hasattr(args, 'main'):
            raise CommandError(f'Wrong command. Supported commands: {[command.name for command in commands]}')
        args.main(args)
    except (CommandError, ConfigurationError) as e:
        logger.exception(f'Start up failed')



from typing import Dict, Tuple, Union, Any


def search_in_tuple(key, value, t: Tuple) -> Union[Dict[str, Any], None]:
    for e in t:
        if type(e) is dict and key in e and e[key] == value:
            return e
    return None






from logging import getLogger
from .utils.serialization import deserializable_exception

logger = getLogger(__name__)


class RemoteError(Exception):
    pass


class FileTypeNotSupported(Exception):
    pass


class DatabaseUriNotFound(Exception):
    pass


class ConfigurationError(Exception):
    pass


class CommandError(Exception):
    pass


class ConsumingTimeout(Exception):
    pass


class NoServiceContainersFound(Exception):
    pass


class WrongServiceContainerType(Exception):
    pass


class NotEnoughWorkers(BaseException):
    pass


class RpcMethodCallTimeout(Exception):
    pass


class ConsumingConnectionError(Exception):
    pass


class BadRequest(Exception):
    pass


class RpcMethodNotFound(Exception):
    pass


class InternalError(Exception):
    pass


class PublishingError(Exception):
    pass


class UnknownService(Exception):
    def __init__(self, service_name):
        self.msg = f'Unknown service `{self.service_name}`'
        super().__init__(self.msg)

    def __str__(self):
        return self.msg


class UnserializableValueError(Exception):
    def __init__(self, value):
        try:
            self.repr_value = repr(value)
        except:
            logger.exception('Repr failed in UnserializableValueError')
            self.repr_value = '[__repr__ failed]'
        self.msg = f'Unserializable value: `{self.repr_value}`'
        super().__init__(self.msg)

    def __str__(self):
        return self.msg


@deserializable_exception
class MethodNotFound(BadRequest):
    pass


@deserializable_exception
class IncorrectSignature(BadRequest):
    pass


@deserializable_exception
class AccessDenied(Exception):
    pass


@deserializable_exception
class RpcSchemaError(BadRequest):
    def __init__(self, message=None):
        if not message:
            message = 'RPC validation failed'
        super().__init__(message)



from multiprocessing import cpu_count
from ..utils.config import update_flat_config_object


class Config:

    default_workers_cpu_based = cpu_count() + 1

    external_conf_file_section_name = 'muhomor'
    rpc_entry_point_attr = 'rpc_entry_point'
    rpc_methods_injected_attr = 'rpc_methods'
    rpc_caller_injected_attr = 'rpc'

    def __init__(self):
        self.amqp_uri: str = 'amqp://guest:guest@localhost:5672//'
        self.heartbeat: int = 11
        self.heartbeat_check: int = 2
        self.workers: int = 0  # set to 0 to use recommended settings, see  `default_workers_cpu_based` attr
        self.restart_dead_workers: bool = False
        self.amqp_prefix: str = 'muhomor'
        self.serializer: str = 'json'
        self.rpc_response_timeout: int = 30
        self.rpc_response_client_pool_timeout: float = 0.1
        self.rpc_reply_queue_ttl: int = 30000  # ms
        self.connect_timeout = 30
        self.transport_options = {
            'max_retries': 60,  # Max num of retries before we give up
            'interval_start': 1,  # How long we start sleeping between retries (sec)
            'interval_step': 1,  # By how much the interval is increased for each retry
            'interval_max': 5,  # Maximum number of seconds to sleep between retries
            'timeout': 5*60  # Maximum seconds waiting before we give up
        }
        self.amqp_ssl = False
        self.max_service_name_len = 30


class BrokerTemplate:

    def __init__(self, global_config: Config):
        self.PREFIX_RPC = f'{global_config.amqp_prefix}.rpc'
        self.PREFIX_RPC_REPLY = f'{global_config.amqp_prefix}.rpc.reply'
        self.PREFIX_EVENT = f'{global_config.amqp_prefix}.evt'

        self.EXCHANGE_NAME_RPC = self.PREFIX_RPC
        self.EXCHANGE_NAME_RPC_REPLY = self.PREFIX_RPC_REPLY
        self.EXCHANGE_NAME_PREFIX_EVENT = self.PREFIX_EVENT
        self.EXCHANGE_NAME_EVT = '{prefix}.{body}'.format(
            prefix=self.EXCHANGE_NAME_PREFIX_EVENT,
            body='{producer}')

        # exclusive: true, auto-delete: true
        self.QUEUE_TEMPLATE_BROADCAST_EVT = '{prefix}|{body}'.format(
            prefix=self.PREFIX_EVENT,
            body='{producer}:{event}(b)->{consumer}:{consumer_method}?{rand}')
        # durable: true
        self.QUEUE_TEMPLATE_SINGLETON_EVT = '{prefix}|{body}'.format(
            prefix=self.PREFIX_EVENT,
            body='{producer}:{event}(s)')
        # durable: true
        self.QUEUE_TEMPLATE_SERVICE_POOL_EVT = '{prefix}|{body}'.format(
            prefix=self.PREFIX_EVENT,
            body='{producer}:{event}(p)->{consumer}:{consumer_method}')
        self.QUEUE_TEMPLATE_RPC = '{prefix}|{body}'.format(
            prefix=self.PREFIX_RPC,
            body='{producer}')
        self.QUEUE_TEMPLATE_RPC_REPLY_PREFIX = f'{self.PREFIX_RPC}.reply'
        self.QUEUE_TEMPLATE_RPC_REPLY = '{prefix}|{body}'.format(
            prefix=self.QUEUE_TEMPLATE_RPC_REPLY_PREFIX,
            body='{producer}:{method}->{consumer}?{rand}')


config = Config()
broker_template = BrokerTemplate(config)


def get_config() -> Config:
    return config


def get_broker_template() -> BrokerTemplate:
    return broker_template


def configure(**kwargs):
    global broker_template
    update_flat_config_object(config, **kwargs)
    broker_template = BrokerTemplate(config)



from re import sub


def replace_non_ascii(value: str, replace: str = '') -> str:
    if isinstance(value, str):
        return sub(r'[^\x00-\x7F]+', replace, value).strip() if len(value) else value
    elif value is None:
        return None
    else:
        raise Exception(f'Wrong type "{type(value)}" for replace_non_ascii. It must be a string.')



from collections import Iterable
from logging import getLogger
from inspect import getmodule
from ..registry import Registry

logger = getLogger(__name__)


def _get_module_path(exc_type):
    module = getmodule(exc_type)
    return f'{module.__name__}.{exc_type.__name__}'


def serialize_safe(value):
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        return {
            serialize_safe(key): serialize_safe(value[key])
            for key in value
        }
    if isinstance(value, Iterable):
        return list(map(serialize_safe, value))
    try:
        return value.__str__()
    except:
        logger.exception('Serialization failed')
        return '[serialization failed]'


def serialize_exception(exc):
    return {
        'exc_type': type(exc).__name__,
        'exc_path': _get_module_path(type(exc)),
        'exc_args': list(map(serialize_safe, exc.args)) if hasattr(exc, 'args') else None,
        'value': serialize_safe(exc),
    }


def deserialize_exception(obj, unserializable_exc_cls = Exception):
    """ Deserialize `obj` to an exception instance.

    If the `exc_path` value matches an exception registered as
    `deserializable_exception`, return an instance of that exception type.
    Otherwise, return an instance of `unserializable_exc_cls` describing the exception.
    """
    exc_type = 'Remote RPC unknown error. Data: {data}'
    value = ''
    if is_serialized_exception(obj):
        key = obj.get('exc_path')
        if key in Registry.exceptions:
            exc_args = obj.get('exc_args', ())
            return Registry.exceptions[key](*exc_args)
        exc_type = obj.get('exc_type')
        value = obj.get('value')
    else:
        try:
            exc_type = exc_type.format(data=obj)
        except:
            logger.exception('Formatting unserializable exception type failed.')
            exc_type = exc_type.format(data='?')
    return unserializable_exc_cls(f'{exc_type}{"" if not len(value) else " " + value}')


def deserializable_exception(exc_type):
    """ Decorator that registers `exc_type` as deserializable back into an
    instance, rather than a :class:`unserializable_exc_cls`. See :func:`deserialize_exception`.
    """
    key = _get_module_path(exc_type)
    Registry.exceptions[key] = exc_type
    return exc_type


def is_serialized_exception(obj) -> bool:
    return obj and type(obj) == dict and \
           'exc_path' in obj and 'exc_args' in obj and 'exc_type' in obj and 'value' in obj



from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import DateTime
from typing import Any
from sqlalchemy.orm.query import Query


class UtcNow(expression.FunctionElement):
    type = DateTime()


@compiles(UtcNow, 'postgresql')
def _pg_utcnow(element, compiler, **kw):
    # return "TIMEZONE('utc', CURRENT_TIMESTAMP)"
    return "(date_part('epoch'::text, now()) * (1000)::double precision)"


@compiles(UtcNow, 'mysql')
def _mysql_utcnow(element, compiler, **kw):
    return "UNIX_TIMESTAMP()"


def compiled_sql(query: Any):
    if isinstance(query, Query):
        return str(query.statement.compile(compile_kwargs={"literal_binds": True}))
    return ''



class Command:
    name = None

    @staticmethod
    def initialize_parser(parser):
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def main(args):
        raise NotImplementedError  # pragma: no cover


# class ShowConfig(Command):
#     """ Output as YAML string the configuration that would be passed to a
#     service.
#
#     Useful for viewing config files that load values from environement
#     variables.
#     """
#
#     name = 'show-config'
#
#     @staticmethod
#     def initialize_parser(parser):
#         parser.add_argument('--config', default='config.yml', help='The YAML configuration file')
#         return parser
#
#     @staticmethod
#     def main(args):
#         from .show_config import main
#         main(args)


class Start(Command):
    """Run nameko services.  Given a python path to a module containing one or
    more nameko services, will host and run them. By default this will try to
    find classes that look like services (anything with nameko entrypoints),
    but a specific service can be specified via
    ``nameko run module:ServiceClass``.
    """

    name = 'start'

    @staticmethod
    def initialize_parser(parser):
        parser.add_argument('path', help='Python path to one or more service classes to run')
        parser.add_argument('--config', default='', help='The YAML configuration file')
        parser.add_argument('--rabbit', default='pyamqp://guest:guest@localhost', help='RabbitMQ url')
        return parser

    @staticmethod
    def main(args):
        from .rpc.cluster import main
        main(args)


class Shell(Command):
    """Launch an interactive python shell for working with remote nameko
    services.

    This is a regular interactive interpreter, with a special module ``n``
    added to the built-in namespace, providing ``n.rpc`` and
    ``n.dispatch_event``.
    """

    name = 'shell'

    SHELLS = ['bpython', 'ipython', 'plain']

    @classmethod
    def initialize_parser(cls, parser):
        parser.add_argument(
            '--broker', default='pyamqp://guest:guest@localhost',
            help='RabbitMQ broker url')
        parser.add_argument(
            '--interface', choices=cls.SHELLS,
            help='Specify an interactive interpreter interface.'
                 ' (Ignored if not in TTY mode)')
        parser.add_argument(
            '--config', default='',
            help='The YAML configuration file')
        return parser

    @staticmethod
    def main(args):
        from .shell import main
        main(args)


commands = Command.__subclasses__()



import sys
from importlib import import_module
import yaml
import pathlib
from typing import Callable
from ..exceptions import FileTypeNotSupported
import os
from logging import getLogger
import json
import pkgutil

logger = getLogger(__name__)


def reraise_import_module_exception(module_name, e):
    missing_module_re = f"^No module named '?{module_name}'?$"
    if re.match(missing_module_re, str(e)):
        raise Exception(f'Module {module_name} not found')
    raise e


def try_to_import_module_by_name(module_name):
    parts = module_name.split(":", 1)
    success = True
    module = None
    if len(parts) == 1:
        module_name, obj = module_name, None
    else:
        module_name, obj = parts[0], parts[1]
    try:
        import_module(module_name)
        module = sys.modules[module_name]
    except ImportError as e:
        # reraise_import_module_exception(module_name, e)
        logger.exception(f'Error importing module `{module_name}`: {e}')
        success = False
    return success, module_name, module, obj


def decorator_containers_from_module(module_name, container_extractor: Callable, **kwargs):
    _is_submodule: bool = kwargs['_is_submodule'] if kwargs and '_is_submodule' in kwargs else False
    containers = list()
    module_loaded, module_name, module, obj = try_to_import_module_by_name(module_name)
    if module_loaded:
        if obj is None:
            containers.extend(container_extractor(module))
            if not len(containers) and not _is_submodule:
                logger.info(f'Decorator containers not found in module `{module_name}`. '
                            'Trying to search in submodules...')
                for importer, modname, ispkg in pkgutil.iter_modules(module.__path__):
                    if not ispkg:
                        logger.info(f'Inspecting submodule `{module_name}.{modname}`')
                        containers_sub = decorator_containers_from_module(f'{module_name}.{modname}',
                                                                          container_extractor,
                                                                          _is_submodule=True)
                        logger.info(f'Found {len(containers_sub)} decorator container(s) '
                                    f'in submodule `{module_name}.{modname}`.')
                        containers.extend(containers_sub)
                    else:
                        logger.debug(f'Submodule `{module_name}.{modname}` skipped, it is package')
        else:
            try:
                container = getattr(module, obj)
                if not isinstance(container, type):
                    logger.error('Decorator container must be a class. '
                                 f'Instead, object `{module_name}:{obj}` '
                                 f'of type {type(container).__name__} was found.')
                else:
                    containers.append(container)
            except AttributeError:
                logger.exception(f'Failed to find decorator container {obj} in module {module_name}.')
    return containers


def load_config(path: str = 'config.yml'):
    ext = (pathlib.Path(path).suffixes[-1]).lower()
    if ext == '.yml' or ext == '.yaml':
        return load_yaml_file(path)
    elif ext == '.json':
        return load_json_file(path)
    raise FileTypeNotSupported(f'Config file of type `{ext}` is not supported.')


def load_yaml_file(path):
    if os.path.exists(path):
        with open(path, 'rt') as f:
            return yaml.full_load(f)
    raise FileNotFoundError(f'Config file of type `{path}` not found.')


def load_json_file(path):
    if os.path.exists(path):
        with open(path, 'rt') as f:
            return json.load(f)
    raise FileNotFoundError(f'Config file of type `{path}` not found.')


def load_config_section(path: str, config_section: str):
    extra_config = load_config(path)
    if config_section in extra_config:
        return extra_config[config_section]
    return None



from typing import Dict, Union, Any
from sqlalchemy.ext.declarative import DeclarativeMeta
from ..exceptions import DatabaseUriNotFound


def update_flat_config_object(config, **kwargs):
    for k in kwargs:
        if hasattr(config, k):
            setattr(config, k, kwargs.get(k))


def database_uri_from_dictionary(config: Dict[str, Any],
                                 prefix: str,
                                 base: Union[DeclarativeMeta, str] = None,
                                 prefix_base_separator: str = ':',
                                 config_key_name: str = 'databases'):
    db_uri = None
    base_name = base.__name__ if type(base) is DeclarativeMeta else base
    uri_key = f'{prefix}{prefix_base_separator}{base_name}' if base else prefix
    db_uris = config.get(config_key_name, None)
    if db_uris:
        db_uri = db_uris.get(uri_key, None)
    if db_uri:
        return db_uri
    raise DatabaseUriNotFound(f'Database URI with key `config->{config_key_name}->{uri_key}` not found in provided config.')

# def load_relative(filename: str = 'config.yml', path: str = '.'):
#     conf_path = Path(sys.argv[0]).parents[len(path)]
#     # curr_path = path.dirname(path.abspath(__file__))
#     # project_path = path.dirname(app_path)
#     config_file_path = path.join(conf_path, filename)
#     config = load_config(config_file_path)
#     return config





import multiprocessing as mp
from multiprocessing import Value, Queue
from typing import Dict, List, Union, Any
import os
import time
import queue
from .config import Config, configure, get_config
from .worker import RpcServer
from .entry_points import _is_entrypoint
from ..exceptions import NotEnoughWorkers
import threading
import inspect
from logging import getLogger
from ..utils.filesystem import load_config_section, decorator_containers_from_module
import sys


logger = getLogger(__name__)


class Cluster:

    def __init__(self, service, workers_num, restart_dead_workers=True):
        self.workers_num = workers_num
        self.service_name = service.name
        self.service_class = service
        self.start_queue = Queue(maxsize=1)
        self.workers = {}
        self.worker_stop_sig: Dict[int, Value] = {}
        self.restart_dead_workers = restart_dead_workers
        self.last_health_report = time.time()
        self.loop_interval = 1
        self._processes_counter = 0

    def run(self):
        logger.info(f'Starting cluster `{self.service_name}`; process id: {os.getpid()}')
        try:
            for _ in range(self.workers_num):
                self.worker_process()
            logger.info(f'Cluster `{self.service_name}` started with {len(self.workers)} workers')
            if len(self.workers) < self.workers_num:
                logger.info(f'Not enough workers. Stopping cluster `{self.service_name}`...')
                raise NotEnoughWorkers
            self.loop(interval=self.loop_interval)
        except KeyboardInterrupt:
            logger.info(f'Stopping cluster `{self.service_name}`')
        except NotEnoughWorkers:
            pass
        finally:
            if not len(self.workers):
                logger.info(f'Cluster `{self.service_name}` stopped because workers pool is empty')
            else:
                for c in self.worker_stop_sig.values():
                    c.value = 1
                for pid in self.workers:
                    self.workers[pid].join()
                logger.info(f'Cluster `{self.service_name}` stopped because all workers completed jobs')

    def start_worker(self, start_queue: Queue, process_stopper: Value, **kwargs):
        try:
            worker = RpcServer(**kwargs)
            worker.setup_provider(self.service_class)
            worker.set_process_stopper(process_stopper)
            start_queue.put(worker.worker_id, block=True, timeout=10)
            worker.run()
        except BaseException as e:
            logger.error(f'Worker creation in cluster `{self.service_name}` failed: {e}')
            start_queue.put(False, block=True, timeout=10)

    def worker_process(self):
        stopper = Value('i', 0)
        self._processes_counter += 1
        proc_name = f'{self.service_name:15s} {self._processes_counter:5d}'
        p = mp.Process(target=self.start_worker,
                       name=f'{proc_name}',
                       daemon=True,
                       args=(self.start_queue, stopper,),
                       kwargs={'name': self.service_name})
        p.start()
        started = False
        try:
            started = self.start_queue.get(block=True, timeout=10)
        except queue.Empty:
            logger.warning(f'Worker start queue is empty in cluster `{self.service_name}`...')
            pass
        if started:
            logger.info(f'{RpcServer.alias(self.service_name, started)} started')
            self.worker_stop_sig[p.pid] = stopper
            self.workers[p.pid] = p
        else:
            p.terminate()

    def loop(self, interval=1):
        while True and len(self.workers) > 0:
            time.sleep(interval)
            dead_workers = list()
            for pid in self.workers:
                if not self.workers[pid].is_alive():
                    dead_workers.append(pid)
            if len(dead_workers) > 0:
                for pid in dead_workers:
                    exitcode = self.workers[pid].exitcode
                    del self.workers[pid]
                    if exitcode != 0:
                        logger.error(f'{RpcServer.alias(self.service_name, pid)} died, code: {exitcode}')
                        if self.restart_dead_workers:
                            logger.info(f'Attempting to start new worker in cluster `{self.service_name}`...')
                            self.worker_process()
                    else:
                        logger.info(f'{RpcServer.alias(self.service_name, pid)} exited.')
            if time.time() - self.last_health_report > 5:
                self.last_health_report = time.time()
                if self.workers_num != len(self.workers):
                    logger.debug(f'Alive workers in cluster `{self.service_name}`: {len(self.workers)}')


def _verify_rpc_provider_class(obj):
    config = get_config()
    if hasattr(obj, 'name'):
        if len(obj.name) > config.max_service_name_len:
            logger.warning(f'Name of RPC provider `{obj.__name__}` exceeds '
                           f'{config.max_service_name_len} chars.')
        else:
            return True
    else:
        logger.warning(f'Class `{obj.__name__}` contains RPC entry points, '
                       'but has no `name` attribute. If you want it to be registered as service, '
                       'include `name` attribute outside of __init__ scope.')
    return False


def rpc_entrypoint_container_extractor(module):
    containers = list()
    for _, maybe_container in inspect.getmembers(module, lambda obj: isinstance(obj, type)):
        if inspect.getmembers(maybe_container, _is_entrypoint):
            if _verify_rpc_provider_class(maybe_container):
                containers.append(maybe_container)
    return containers


def _run_one(service):
    config = get_config()
    cluster = Cluster(service=service,
                      workers_num=config.workers if config.workers > 0 else config.default_workers_cpu_based,
                      restart_dead_workers=config.restart_dead_workers)
    cluster.run()


def _run_multi(services):
    threads = []
    for service in services:
        t = threading.Thread(target=_run_one, daemon=True, args=(service,))
        t.start()
        threads.append(t)
    while True:
        try:
            time.sleep(5)
        except KeyboardInterrupt:
            time.sleep(5)
            break


def run_cluster(service_or_services: Union[str, List[type], type], conf: Dict[str, Any] = None):
    """
    Starting cluster of services with multiple workers per service.
    Each cluster runs in separate thread. And every of those threads creates separate process for each service worker.
    If you run cluster on MacOS it is possible that you will face fork problem. Details are here:
    http://sealiesoftware.com/blog/archive/2017/6/5/Objective-C_and_fork_in_macOS_1013.html
    To fix it, add following to .bash_profile:
    export OBJC_DISABLE_INITIALIZE_FORK_SAFETY=YES
    """
    if type(service_or_services) is str and \
            (service_or_services.__contains__('.') or service_or_services.__contains__(':')):
        services = decorator_containers_from_module(service_or_services, rpc_entrypoint_container_extractor)
        if services and len(services):
            return run_cluster(services, conf)
        else:
            logger.critical('Running cluster failed. No RPC providers found.')
            return
    services = [service_or_services] if isinstance(service_or_services, type) else service_or_services
    if conf is not None:
        configure(**conf)
    if len(services) > 1:
        _run_multi(services)
    else:
        _run_one(services[0])


def main(args):
    if '.' not in sys.path:
        sys.path.insert(0, '.')

    if not args.path or not len(args.path):
        logger.critical(f'Critical error! Wrong services path: `{args.path}`')
        return

    run_cluster(args.path,
                load_config_section(args.config, Config.external_conf_file_section_name) if args.config else None)



from .config import get_config, get_broker_template
from kombu import Connection, Queue, Exchange, Consumer, Producer
from typing import Dict, Any, Union
import socket
from kombu.message import Message
from ..exceptions import RpcMethodCallTimeout, PublishingError
from logging import getLogger

logger = getLogger(__name__)

rpc_exchange: Exchange = None
evt_exchange: Dict[str, Exchange] = {}


def get_rpc_exchange() -> Exchange:
    global rpc_exchange
    if rpc_exchange is None:
        rpc_exchange = Exchange(get_broker_template().EXCHANGE_NAME_RPC,
                                type='topic',
                                durable=True,
                                delivery_mode=Exchange.PERSISTENT_DELIVERY_MODE,
                                auto_delete=False)
    return rpc_exchange


def get_evt_exchange(service_name: str) -> Exchange:
    global evt_exchange
    if service_name not in evt_exchange:
        evt_exchange[service_name] = Exchange(get_broker_template().EXCHANGE_NAME_EVT.format(producer=service_name),
                                              type='topic',
                                              durable=True,
                                              delivery_mode=Exchange.PERSISTENT_DELIVERY_MODE,
                                              auto_delete=True)
    return evt_exchange[service_name]


def make_connection(heartbeats_on: bool = True) -> Connection:
    config = get_config()
    logger.debug(f'Making connection in make_connection, config: {config.amqp_uri}')
    logger.debug(f'Making connection in make_connection, config func: {get_config().amqp_uri}')
    c = Connection(config.amqp_uri,
                      heartbeat=config.heartbeat if heartbeats_on else 0,
                      connect_timeout=config.connect_timeout,
                      transport_options=config.transport_options)
    logger.debug(f'Created  connection: {c}')
    return c


def publish_rpc_call(service_name: str, method_name: str,
                     args: Union[Dict[str, Any], None],
                     producer: Producer,
                     consumer: Consumer,
                     reply_queue: Queue,
                     correlation_id: str):
    config = get_config()
    try:
        consumer.add_queue(reply_queue)
        consumer.consume(no_ack=True)
        producer.publish(args,
                         exchange=get_rpc_exchange(),
                         routing_key=f'{service_name}.{method_name}',
                         # declare=[reply_queue],
                         reply_to=reply_queue.name,
                         correlation_id=correlation_id)
        consumer.connection.drain_events(timeout=config.rpc_response_timeout)
        consumer.cancel()
        consumer.queues = {}
    except socket.timeout as e:
        raise RpcMethodCallTimeout
    except BaseException as e:
        logger.warning(f'Rpc call publishing failed: {e}')
        raise PublishingError


def rpc_queue(name: str, exchange: Exchange, routing_key: str) -> Queue:
    return Queue(
        name,
        exchange=exchange,
        routing_key=routing_key,
        durable=True,
        message_ttl=30,
        # expires=5000,
        auto_delete=False
    )


def evt_queue(name: str, exchange: Exchange, routing_key: str, binding_arguments: Dict[str, Any] = None) -> Queue:
    return Queue(
        name,
        exchange=exchange,
        routing_key=routing_key,
        durable=True,
        message_ttl=10,
        # expires=5000,
        auto_delete=False,
        binding_arguments=binding_arguments
    )


def is_rpc_exchange(exchange_name: str) -> bool:
    return exchange_name == get_broker_template().EXCHANGE_NAME_RPC


def is_rpc_reply_message(message: Message, for_service: str) -> bool:
    rk = message.delivery_info['routing_key']
    if message.delivery_info['exchange'] == '' and rk.startswith(get_broker_template().QUEUE_TEMPLATE_RPC_REPLY_PREFIX):
        return rk.split('->')[-1].startswith(f'{for_service}?')
    return False


def is_events_exchange(exchange_name: str) -> bool:
    return exchange_name.startswith(get_broker_template().EXCHANGE_NAME_PREFIX_EVENT)


def message_requires_reply(message: Message) -> bool:
    return message and hasattr(message, 'properties') and \
           'reply_to' in message.properties and \
           'correlation_id' in message.properties



from typing import Dict, Any, Union, List
from functools import partial
import types
import jsonschema
from ..exceptions import RpcSchemaError
import inspect
from .config import Config
import os


class RpcEntryPoint:

    def __new__(cls, *args, **kwargs):
        inst = super().__new__(cls)
        inst.__params = (args, kwargs)
        return inst

    def __init__(self, name: str, **kwargs):
        self.pid = os.getpid()
        self.name = name
        self.method_name = None
        self.method_container = None
        self.validation_schema = kwargs['schema'] if 'schema' in kwargs else None
        self.validation_error_code = None
        self.__params = None

    def call(self, params: Union[Dict[str, Any], List[Any]] = None):
        args = []
        kwargs = {}
        if params:
            if type(params) is list:
                args = params
            else:
                kwargs = params

        # logger.debug(f'RPC entry `{self.method_container.name}.{self.name}` <PID {self.pid}> called with '
        #              f'args <{args}> and kwargs <{kwargs}>.')

        if self.validation_schema:
            if len(args):
                validate_json_rpc_data(list(args), self.validation_schema, RpcSchemaError)
            elif len(kwargs):
                validate_json_rpc_data(kwargs, self.validation_schema, RpcSchemaError)
        self.check_signature(args, kwargs)
        return getattr(self.method_container, self.method_name)(*args, **kwargs)

    def bind(self, method_container, method_name):
        if not hasattr(method_container, Config.rpc_methods_injected_attr):
            raise RuntimeError(f'Cannot bind RPC method `{method_name}`. '
                               f'Class `{method_container.__class__.__name__}` must have '
                               f'`{Config.rpc_methods_injected_attr}<dict>` attribute.')
        if self.is_bound():
            raise RuntimeError(f'Cannot bind already bound entry point `{method_name}` as `{self.name}`.')
        instance = self
        instance.method_name = method_name
        instance.method_container = method_container
        container_rpc_methods = getattr(instance.method_container, Config.rpc_methods_injected_attr)
        container_rpc_methods[self.name] = instance
        instance.pid = os.getpid()
        return instance

    def is_bound(self):
        return self.method_container is not None

    def check_signature(self, args, kwargs):
        fn = getattr(self.method_container, self.method_name)
        try:
            inspect.getcallargs(fn, *args, **kwargs)
        except TypeError as exc:
            raise RpcSchemaError(f'Wrong method signature. {str(exc)}')
        except Exception as exc:
            raise RpcSchemaError(f'Checking method signature failed. {str(exc)}')

    @classmethod
    def rpc_entry_point(cls, *decoargs, **decokwargs):
        """
        Decorator for RPC methods. Only instance methods can be decorated.

        :param decoargs: If first argument is provided, it is used as method name.
        :param decokwargs: If 'name' key is provided, it is used as method name.
        :return:
        """

        def wrapped_f(fn, args, kwargs):
            name = decokwargs.get('name', fn.__name__)
            if not (len(decoargs) == 1 and isinstance(decoargs[0], types.FunctionType)):
                if len(args) >= 1:
                    name = args[0]
                    args = args[1:]

            instance = cls(name, *args, **kwargs)
            setattr(fn, Config.rpc_entry_point_attr, instance)
            return fn

        if len(decoargs) == 1 and isinstance(decoargs[0], types.FunctionType):
            return wrapped_f(decoargs[0], args=(), kwargs={})
        else:
            return partial(wrapped_f, args=decoargs, kwargs=decokwargs)


rpc_entry_point = RpcEntryPoint.rpc_entry_point


def _is_entrypoint(method):
    return hasattr(method, Config.rpc_entry_point_attr)


def validate_json_rpc_data(data: Any, schema: Any, exception_cls: type = RpcSchemaError):
    try:
        jsonschema.validate(data, schema, format_checker=jsonschema.FormatChecker())
    except jsonschema.ValidationError as e:
        raise exception_cls(e.message)


def get_entrypoint(method):
    if _is_entrypoint(method):
        return getattr(method, Config.rpc_entry_point_attr, None)
    return None



from sqlalchemy.dialects.mysql import TINYBLOB
from sqlalchemy import func


class TINYBLOB_ENCODED(TINYBLOB):
    """MySQL TINYBLOB type, for binary data up encoded using ENCODE() func."""

    def __init__(self, secret: str, length: int = None):
        TINYBLOB.__init__(self, length=length)
        self.secret = secret

    def bind_expression(self, bindvalue):
        return func.ENCODE(bindvalue, self.secret)

    def column_expression(self, col):
        return func.DECODE(col, self.secret)




from multiprocessing import Value
from kombu import Queue, uuid, Consumer, Producer
import sys
import queue
from typing import Dict, Any, Union, List
import threading
import os
import time
import socket
from .config import Config, BrokerTemplate, get_config, get_broker_template
from .transport import (
    rpc_queue,
    evt_queue,
    is_rpc_exchange,
    is_rpc_reply_message, is_events_exchange,
    message_requires_reply
)
from ..utils.serialization import serialize_exception, serialize_safe as safe_exc_serialize, deserialize_exception
from ..exceptions import (
    UnserializableValueError, IncorrectSignature, RpcMethodNotFound, InternalError, AccessDenied, RemoteError
)
import inspect
from functools import partial
import types
from .transport import get_rpc_exchange, get_evt_exchange, make_connection, publish_rpc_call
from .entry_points import RpcEntryPoint, _is_entrypoint, get_entrypoint
from kombu.utils.debug import setup_logging
from logging import getLogger

logger = getLogger(__name__)

# setup_logging(loglevel=logging.DEBUG, loggers=['kombu.connection', 'kombu.channel'])


W_CONN_LOST = 'Connection to broker lost, trying to re-establish connection...'


class RpcReply:
    def __init__(self, result=None, exception=None, correlation_id=None):
        self.result: Any = result
        self.exception: Any = exception
        self.correlation_id: Union[None, str] = correlation_id
        self._json_response = {'result': None, 'error': None}

    @property
    def json_response(self):
        if self.result is not None:
            try:
                # self._json_response['result'] = kombu_serialization.dumps(self.result, config.serializer)[-1]
                self._json_response['result'] = self.result
            except Exception:
                self.exception = UnserializableValueError(self.result)
                self._json_response['result'] = None
        if self.exception is not None:
            f = serialize_exception if isinstance(self.exception, BaseException) else safe_exc_serialize
            self._json_response['error'] = f(self.exception)
        return self._json_response


class WorkerBase:

    def __init__(self, name: str, heartbeats_on: bool = True):
        self.name = name
        self.worker_id = os.getpid()
        self.worker_alias = WorkerBase.alias(self.name, self.worker_id)
        self.config: Config = get_config()
        self.broker_template: BrokerTemplate = get_broker_template()
        logger.debug(f'Making connection in WorkerBase, config: {self.config.amqp_uri}')
        self.connection = make_connection(heartbeats_on)
        self.rpc_exchange = get_rpc_exchange()
        self.incoming_message_log_on = False
        self.producer = Producer(self.connection, serializer=self.config.serializer)
        self._correlation_id = None

    @staticmethod
    def alias(service_name, worker_id):
        return f'Worker::{service_name}<{worker_id}>'

    @property
    def next_correlation_id(self):
        self._correlation_id = uuid()
        return self._correlation_id

    @property
    def callbacks(self):
        callbacks = list()
        if self.incoming_message_log_on:
            callbacks.append(self.log_incoming_message)
        callbacks.append(self.on_message)
        return callbacks

    def on_message(self, body, message):
        raise NotImplementedError('Subclass must implement this functionality')

    def log_incoming_message(self, body, message):
        if is_rpc_exchange(message.delivery_info["exchange"]):
            logger.debug(f'RpcRequest {message.delivery_info["exchange"]}:{message.delivery_info["routing_key"]}')
        elif is_rpc_reply_message(message, self.name):
            logger.debug(f'RpcReply {message.delivery_info}')
        elif is_events_exchange(message.delivery_info["exchange"]):
            logger.debug(f'Event {message.delivery_info["exchange"]}:{message.delivery_info["routing_key"]}')

    def should_stop(self) -> bool:
        raise NotImplementedError('Subclass must implement this functionality')

    def stop(self):
        logger.debug(f'Worker is stopping...')
        try:
            self.connection.release()
        except BaseException as e:
            logger.warn(f'Error releasing connection: {e}')
        logger.info(f'Worker stopped.')

    def heartbeat(self, frequency: int = 1):
        while not self.should_stop():
            try:
                self.connection.heartbeat_check()
            except BaseException as e:
                logger.warning(f'Heartbeat failed: {e}')
                pass
            time.sleep(frequency)
        logger.info(f'Heartbeats stopped.')

    def start_heartbeats(self):
        if not self.should_stop():
            threading.Thread(target=self.heartbeat, args=(self.config.heartbeat_check,), daemon=True).start()


class RpcWorker(WorkerBase):

    def __init__(self, name: str, heartbeats_on: bool = True):
        WorkerBase.__init__(self, name, heartbeats_on=heartbeats_on)
        self._heartbeat_sig = queue.Queue()
        self._process_stop = None
        self._stopping = False
        self._consuming_channel = None
        self._producing_channel = None
        self.queues = list()

    def set_process_stopper(self, val: Union[Value, None]):
        self._process_stop = val

    def should_stop(self) -> bool:
        return self._stopping or not (self._process_stop is None or self._process_stop.value == 0)

    def on_message(self, body, message):
        raise NotImplementedError('Subclass must implement this functionality')


class RpcCallPublisher(WorkerBase):

    def _setup_publisher(self):
        self.rpc_response_client_pool = queue.Queue()
        self.queues_for_rpc_reply = dict()
        logger.debug(f'Connection in _setup_publisher: {self.connection}')
        self.consumer_rpc_reply = Consumer(self.connection,
                                           callbacks=self.callbacks,
                                           no_ack=False,
                                           auto_declare=True,
                                           accept={'application/json'},
                                           prefetch_count=1)

    def setup_publisher(self):
        self._setup_publisher()

    def _make_queue_for_rpc_reply(self, service: str, method: str) -> Queue:
        reply_queue = Queue(self.broker_template.QUEUE_TEMPLATE_RPC_REPLY.
                            format(producer=service, method=method, consumer=self.name, rand=uuid()),
                            exchange=self.rpc_exchange,
                            exclusive=True,
                            auto_delete=True,
                            no_ack=True,
                            queue_arguments={'x-expires': self.config.rpc_reply_queue_ttl})
        if service not in self.queues_for_rpc_reply:
            self.queues_for_rpc_reply[service] = {}
        self.queues_for_rpc_reply[service][method] = reply_queue
        return reply_queue

    def get_queue_for_rpc_reply(self, service: str, method: str) -> Queue:
        if service in self.queues_for_rpc_reply and method in self.queues_for_rpc_reply[service]:
            return self.queues_for_rpc_reply[service][method]
        return self._make_queue_for_rpc_reply(service, method)

    def should_stop(self) -> bool:
        pass

    def _get_from_rpc_response_client_pool(self):
        return self.rpc_response_client_pool.get(block=True,
                                                 timeout=self.config.rpc_response_client_pool_timeout)

    def rpc(self, service_name, method_name, args: Dict[str, Any] = None):
        publish_rpc_call(service_name, method_name, args,
                         producer=self.producer, consumer=self.consumer_rpc_reply,
                         reply_queue=self.get_queue_for_rpc_reply(service_name, method_name),
                         correlation_id=self.next_correlation_id)
        response = None
        try:
            response = self._get_from_rpc_response_client_pool()
        except queue.Empty:
            pass
        if response and 'error' in response and response['error'] is not None:
            e = deserialize_exception(response['error'], RemoteError)
            if isinstance(e, AccessDenied):
                logger.warning(f'RPC access denied to {service_name}.{method_name}')
            raise e
        return response['result'] if response and 'result' in response else None

    def verify_correlation_id(self, message) -> bool:
        return message and hasattr(message, 'properties') and \
               'correlation_id' in message.properties and \
               message.properties['correlation_id'] == self._correlation_id

    def is_my_rpc_reply_message(self, message) -> bool:
        return is_rpc_reply_message(message, self.name) and self.verify_correlation_id(message)

    def on_message(self, body, message):
        if self.is_my_rpc_reply_message(message):
            self._on_if_my_rpc_reply(body, message)
        message.ack()

    def on_rpc_reply(self, body, message):
        if self.is_my_rpc_reply_message(message):
            self._on_if_my_rpc_reply(body, message)

    def _on_if_my_rpc_reply(self, body, message):
        self.rpc_response_client_pool.put(body)


class RpcClient(RpcWorker, RpcCallPublisher):

    def __init__(self, name: str):
        RpcWorker.__init__(self, name, heartbeats_on=False)
        logger.debug(f'MUHOMOR CONFIG IN RpcClient: {get_config().amqp_uri}')
        logger.debug(f'MUHOMOR CONFIG IN RpcClient PARENT: {self.config.amqp_uri}')
        self.setup_publisher()

    def on_message(self, body, message):
        RpcCallPublisher.on_message(self, body, message)


# TO-DO: test launching when there are still messages in queues
class RpcServer(RpcWorker, RpcCallPublisher):

    def __init__(self, name: str):
        RpcWorker.__init__(self, name, heartbeats_on=True)
        # RpcCallPublisher.__init__(self, name)
        self.rpc_result_server_pool = queue.Queue()
        self.rpc_method_providers = {}
        self.event_handlers = {}
        self.evt_exchange = get_evt_exchange(self.name)
        self.rpc_routing_key = f'{self.name}.*'
        self.evt_routing_key = '*.*'
        self.provider = None

        self._consuming_channel = None
        self._producing_channel = None
        self._pending_rpc_call_message = None

        self.rpc_queue = rpc_queue(self.broker_template.QUEUE_TEMPLATE_RPC.format(producer=self.name),
                                   self.rpc_exchange, self.rpc_routing_key)

        # self.evt_queue = evt_queue(f'evt.{self.name}', self.evt_exchange, self.evt_routing_key,
        #                            binding_arguments={
        #                                'event': 'ping',
        #                                'x-match': 'any'
        #                            })

        self.queues.append(self.rpc_queue)

        self.consumer: Consumer = None

        self.setup()
        self.setup_publisher()

        # self.start_heartbeats()
        logger.info(f'Created worker with URI `{self.connection.as_uri()}`')

        self.rpc_threads = dict()

    def setup(self):
        self.consumer = Consumer(self.connection,
                                 queues=self.queues,
                                 callbacks=self.callbacks,
                                 no_ack=False,
                                 auto_declare=True,
                                 accept={'application/json'},
                                 prefetch_count=1)
        self.consumer.consume()

    def _enter_rpc_entry_point(self, entry_point: RpcEntryPoint, params: Union[Dict[str, Any], List[Any]],
                               correlation_id: str = None, return_result: bool = False) -> Union[RpcReply, None]:
        """
        Call wrapped RPC method, locally.
        :param entry_point: RpcEntryPoint instance created using decorator.
        :param params: Called Rpc method parameters.
        :param correlation_id: If answer is not required, it should be set to None.
        :param return_result: If set to True, result will be returned directly to caller.
                              Otherwise it will be saved in results queue.
        :return:
        """
        reply = RpcReply(correlation_id=correlation_id)
        try:
            reply.result = entry_point.call(params)
        except BaseException as e:
            reply.exception = e
        if return_result:
            return reply
        if correlation_id is not None:
            self.rpc_result_server_pool.put(reply)
        else:
            del reply
        return

    def _set_pending_rpc_call_message(self, message, body):
        message.properties['rpc_result_correlation_id'] = uuid()
        message.properties['queued_at'] = time.time()
        message.properties['body'] = body
        # logger.debug('Setting pending rpc message')
        self._pending_rpc_call_message = message
        return message.properties['rpc_result_correlation_id']

    def _publish_rpc_reply(self, message, reply: RpcReply):
        self._pending_rpc_call_message = None
        try:
            if message_requires_reply(message):
                logger.debug(f'Publishing reply '
                             f'{reply.json_response}')
                self.producer.publish(
                    reply.json_response,
                    exchange='',
                    routing_key=message.properties['reply_to'],
                    correlation_id=message.properties['correlation_id'],
                    serializer=self.config.serializer,
                    retry=True)
        except Exception as e:
            logger.warning(f'Publishing RPC reply failed: {e}', e)

    def _get_rpc_reply(self):
        res = None
        try:
            res = self.rpc_result_server_pool.get_nowait()
        except queue.Empty:
            pass
        except Exception as e:
            logger.warning(f'Getting rpc call result queue error: {e}')
        return res

    def setup_provider(self, provider):
        try:
            self.provider = provider()
            setattr(self.provider, self.config.rpc_methods_injected_attr, dict())
            setattr(self.provider, self.config.rpc_caller_injected_attr, self.rpc)
        except BaseException:
            logger.exception(f'Instantiation of provider in {self.worker_alias} failed.')
            raise
        for method_name, method in inspect.getmembers(provider, _is_entrypoint):
            entrypoint = get_entrypoint(method)
            try:
                entrypoint.bind(self.provider, method_name)
            except BaseException:
                logger.exception(f'Binding RPC provider for {self.worker_alias} failed.')
                raise
        logger.debug(f'PRC provider for {self.worker_alias} is set.')

    def connect(self):
        # revived_connection = self.connection.clone()
        logger.debug(f'Worker {self.worker_id} is connecting')
        try:
            self.connection.ensure_connection(max_retries=3)
            self._consuming_channel = self.connection.default_channel
            self._producing_channel = self.connection.channel()

            self.consumer_rpc_reply.revive(self._consuming_channel)
            self.producer.revive(self._producing_channel)
            self.consumer.revive(self._consuming_channel)
            self.consumer.consume()
        except BaseException:
            logger.exception('Connection failed')

    # TO-DO: implement real services auth functionality
    def validate_rpc_access(self, calling_service: str, method: str) -> Union[bool, BaseException]:
        if calling_service == 'subscriptions':
            if method == 'mint':
                return AccessDenied(f'Service `{calling_service}` has no access to method `{self.name}.{method}`')
        return True

    def on_message(self, body, message):
        if is_rpc_exchange(message.delivery_info["exchange"]):
            self.on_rpc(body, message)
        elif is_rpc_reply_message(message, self.name):
            self.on_rpc_reply(body, message)
        elif is_events_exchange(message.delivery_info["exchange"]):
            self.on_event(body, message)

    def on_rpc(self, body, message):
        provider = self.get_provider_for_method(message.delivery_info['routing_key'])
        corr_id = None

        requesting_service = 'subscriptions'

        valid_access = self.validate_rpc_access(requesting_service, provider.method_name)
        if isinstance(valid_access, BaseException):
            if message_requires_reply(message):
                # corr_id = self._set_pending_rpc_call_message(message)
                logger.error(f'Attempt of illegal access of method {provider.method_name} '
                             f'from service {requesting_service}')
                self._publish_rpc_reply(message, RpcReply(exception=valid_access, correlation_id=uuid()))
        else:
            if message_requires_reply(message):
                # logger.debug('RPC message requires reply... Putting in pending queue...')
                corr_id = self._set_pending_rpc_call_message(message, body)

            t_name = f'EP({provider.method_name}):{len(self.rpc_threads) + 1}'
            t = threading.Thread(target=self._enter_rpc_entry_point,
                                 daemon=True,
                                 name=t_name,
                                 args=(provider, body,),
                                 kwargs={'correlation_id': corr_id})
            # self.rpc_threads[t_name] = t
            t.start()
        message.ack()
        # logger.debug(f'RPC message in pending queue 2222: {self._pending_rpc_call_message}, corrid={corr_id}')

    def on_event(self, body, message):
        producer_name = message.delivery_info["exchange"].split(".")[-1]
        routing_key = message.delivery_info['routing_key']
        try:
            evt_handler_provider = self.get_event_handler(producer_name, routing_key)
            if evt_handler_provider is not None:
                logger.debug(f'Event handler provider: {evt_handler_provider} {message.payload}.')
                if evt_handler_provider:
                    evt_handler_provider.call(**body)
        except BaseException:
            logger.exception('Event handling failed.')
        message.ack()

    def dispatch(self, event_name: str, payload: Any = None, event_sender: str = None):
        self.producer.publish(payload, exchange=self.evt_exchange,
                              routing_key=f'{event_sender or self.name}.{event_name}')

    def check_pending_rpc_call_message(self):
        if self._pending_rpc_call_message is not None:
            queued_at = self._pending_rpc_call_message.properties['queued_at'] or 0
            if time.time() - queued_at >= self.config.rpc_response_timeout:
                logger.warning(f'Pending rpc call message {self._pending_rpc_call_message} expired...')
                self._pending_rpc_call_message = None
            else:
                res: RpcReply = self._get_rpc_reply()
                if res is not None and res.correlation_id and \
                        self._pending_rpc_call_message.properties['rpc_result_correlation_id'] == res.correlation_id:
                    self._publish_rpc_reply(self._pending_rpc_call_message, res)

    @property
    def consume_paused(self):
        return self._pending_rpc_call_message is not None

    def on_consuming_iteration(self):
        pass

    # TO-DO: re-connect in case of failure in heartbeats thread
    def consume(self, timeout=None, safety_interval=1, **kwargs):
        if not self._stopping:
            elapsed = 0
            self.connect()
            while not self.should_stop():
                self.on_consuming_iteration()
                self.check_pending_rpc_call_message()
                if not self.consume_paused:
                    try:
                        self.connection.drain_events(timeout=safety_interval)
                    except socket.timeout as e:
                        elapsed += safety_interval
                        self.connection.heartbeat_check()
                        if timeout and elapsed >= timeout:
                            logger.debug(f'Socket timeout in consuming loop: {e}')
                            # raise ConsumingTimeout(e.__str__())
                            raise
                    except socket.error as e:
                        logger.warning(f'Socket error in consuming loop: {e}')
                        if not self.should_stop():
                            raise
                    else:
                        elapsed = 0

    def run(self, **kwargs):
        while not self.should_stop():
            try:
                # pass
                self.consume(**kwargs)
            except (self.connection.connection_errors + self.connection.channel_errors):
                logger.warning(W_CONN_LOST)
                # logger.warning(W_CONN_LOST, exc_info=1)
            except KeyboardInterrupt:
                self._stopping = True
                break
        self.stop()

    def get_provider_for_method(self, routing_key: str) -> RpcEntryPoint:
        method_name = routing_key.split(".")[-1]
        try:
            rpc_methods = getattr(self.provider, self.config.rpc_methods_injected_attr)
        except BaseException as e:
            logger.error(f'No {self.config.rpc_methods_injected_attr} in provider...')
            raise RpcMethodNotFound(f'Method {method_name} not found')

        if method_name in rpc_methods:
            if not hasattr(rpc_methods[method_name], 'call'):
                raise InternalError(f'Provider of method {method_name} is not callable')
            return rpc_methods[method_name]
        else:
            raise RpcMethodNotFound(f'Method {method_name} not found')

    # def get_event_handler(self, service_name, routing_key):
    #     event_name = routing_key
    #     if evt_handlers and self.name in evt_handlers and service_name in evt_handlers[self.name] and \
    #             event_name in evt_handlers[self.name][service_name]:
    #         if hasattr(evt_handlers[self.name][service_name][event_name], 'call'):
    #             return evt_handlers[self.name][service_name][event_name]
    #         else:
    #             logger.warning(f'Event handler provider of service `{self.name}` '
    #                            f'for event {service_name}.{event_name} is not callable')
    #     return None

# class RpcService:
#
#     def __init__(self, service_caller: Callable, resource: object, connection_releaser: Callable):
#         self.service_caller = service_caller
#         self.resource = resource
#         self.connection_releaser = connection_releaser
#
#     def enable_exclusive_connection(self, value: bool = True):
#         self.resource.exclusive_request_rpc_connection = value
#         if not value:
#             self.connection_releaser()
#
#     def __getattr__(self, name):
#         return self.service_caller(name)



import os
from logging import getLogger
from ..rpc.config import configure
from ..rpc.worker import RpcClient


logger = getLogger(__name__)


class FalconRpcMiddleware:

    def __init__(self, service_name: str, **config):
        if config:
            configure(**config)
        logger.debug(f'Initializing RPC Client, PID {os.getpid()}')
        self.rpc = RpcClient(service_name)

    def process_resource(self, req, resp, resource, params):
        resource.rpc = self.rpc

    def process_response(self, req, resp, resource, req_succeeded):
        if hasattr(resource, 'rpc'):
            pass


